'use strict';
const axios = require('axios');
const crypto = require('crypto');

/**
 * Creates an instance of the JotUrl SDK.
 *
 * @param _username the username used to login into JotURL dashboard
 * @param _public_key public api key, you can find it on https://www.joturl.com/reserved/settings.html#tools-api
 * @param _private_key private api key, you can find it on https://www.joturl.com/reserved/settings.html#tools-api
 * @param base_url the base URL to be use to call API endpoints, defaults to https://joturl.com/a/i1/
 *
 * @returns {JotUrlSDK}
 * @constructor
 */
function JotUrlSDK(_username, _public_key, _private_key, base_url) {
    if (!(this instanceof JotUrlSDK)) {
        return new JotUrlSDK(_username, _public_key, _private_key, base_url);
    }

    let self = this;

    self.username = _username || '';
    self.public_key = _public_key || '';
    self.private_key = _private_key || '';
    self._base_url = base_url || 'https://joturl.com/a/i1/';
    self.session_id = '';

    self.checkCallback = callback => {
        return callback || (() => {
        });
    };

    self.download = (url, postParameters = {}) => {
        return new Promise((callback, reject) => {
            axios.post(url, postParameters).then(response => {
                callback((response || {}).data);
            }).catch(e => {
                reject(e);
            });
        });
    };

    self.gmdate = () => {
        let iso = (new Date()).toISOString();

        return iso.substr(0, iso.length - 8) + 'Z';
    };

    self.create_sha256 = (key, message) => {
        let hmac = crypto.createHmac('sha256', key);
        hmac.update(message);

        return hmac.digest('hex');
    };

    self.getHash = (value, time) => {
        let msg = value + ":" + (time ? time : self.gmdate());

        return self.create_sha256(self.private_key, msg);
    };

    self.merge = (a1, a2) => {
        return Object.assign(a1 || {}, a2 || {});
    };

    self.getSecurityParameters = () => {
        let parameters = {};

        if (self.session_id != '') {
            parameters._sid = self.session_id;
            parameters._h = self.getHash(self.session_id);
        }

        return parameters;
    };

    self.buildQuery = params => {
        let result = '';

        params = params || {};
        for (let k in params) {
            if (result) {
                result += '&';
            }
            result += encodeURIComponent(k) + '=' + encodeURIComponent(params[k]);
        }

        return result;
    };

    self.buildURL = (endpoint, parameters) => {
        parameters = self.buildQuery(self.merge(parameters, self.getSecurityParameters()));

        return self._base_url + endpoint + (parameters ? '?' + parameters : '');
    };

    self.getSecurityParameters = () => {
        let parameters = {};

        if (self.session_id) {
            parameters['_sid'] = self.session_id;
            parameters['_h'] = self.getHash(self.session_id);
        }

        return parameters;
    };

    self.login = (force) => {
        return new Promise((callback, reject) => {
            if (!force && self.session_id) {
                callback();
                return;
            }

            let password = self.getHash(self.public_key);

            //try to login
            let url = self.buildURL('users/login', {'username': self.username, 'password': password});
            self.download(url).then(result => {
                if (typeof result === 'object' && result.result) {
                    let sid = self.getParam(result.result, 'session_id');
                    if (sid) {
                        self.session_id = sid;

                        callback();

                        return;
                    }
                }

                throw "unable to login";
            }).catch(e => {
                reject(e);
            });
        });
    };

    self._call = (endpoint, parameters = {}, postParameters = {}, retry = true, force = false) => {
        return new Promise((callback, reject) => {
            self.login(force).then(() => {
                let url = self.buildURL(endpoint, parameters);
                self.download(url, postParameters).then(json => {
                    if (json) {
                        if (typeof json === 'object') {
                            let status = self.getParam(json, 'status', {});
                            let code = self.getParam(status, 'code');

                            if (code != 200) {
                                throw [
                                    'CODE = ' + self.getParam(status, 'code'),
                                    'STATUS = ' + self.getParam(status, 'text'),
                                    'ERROR = ' + "" + self.getParam(status, 'error'),
                                ].join(' - ');
                            }

                            callback.apply(self, [self.getParam(json, 'result', {})]);
                        } else {
                            //maybe we are calling APIs too fastly
                            throw 'malformed response (mr1)';
                        }
                    } else {
                        //maybe we are calling APIs too fastly
                        throw 'malformed response (mr2)';
                    }
                }).catch(e => {
                    if (retry && e && e.response && e.response.statusText && e.response.statusText.indexOf('INVALID session') >= 0) {
                        self._call(endpoint, parameters, postParameters, false, true).then(callback, reject);
                    } else {
                        reject(e);
                    }
                });
            }).catch(e => {
                reject(e);
            });
        });
    };

    /**
     * Call and get results from the API endpoint.
     *
     * @param endpoint API endpoint to be called
     * @param parameters [OPTIONAL] parameters to be passed to the call
     * @param postParameters [OPTIONAL] post parameters to be passed to the call
     *
     * @returns {Promise}
     */
    self.call = (endpoint, parameters = {}, postParameters = {}) => {
        return self._call(endpoint, parameters, postParameters);
    };

    /**
     * Helper that returns the value corresponding to a key in the object obj.
     *
     * @param obj object from which to extract the value
     * @param key key that identifies the value to extract
     * @param def default value to be returned if $key does not exists in obj
     * @returns {*}
     */
    self.getParam = (obj, key, def) => {
        if (typeof def === "undefined") {
            def = '';
        }

        return typeof obj === 'object' && typeof obj[key] !== 'undefined' ? obj[key] : def;
    };

    /**
     * Set the base URL for API calls.
     *
     * @param url base URL for API calls
     */
    self.setBaseURL = url => {
        self._base_url = url;
    };

    /**
     * Get the base URL currently used for API calls.
     *
     * @param url
     * @returns {string}
     */
    self.getBaseURL = () => {
        return self._base_url;
    };
};

module.exports = JotUrlSDK;
module.exports.JotUrlSDK = JotUrlSDK;
module.exports.default = JotUrlSDK;